# -*- coding: utf-8 -*-
"""
Created on Mon Dec 25 22:49:11 2023

@author: Ahti
"""

from scipy.stats import chi2_contingency
import numpy as np
from scipy import stats
from datetime import datetime
from Read_KS_korp_analyys_data import extract_frequencies_from_file_to_dict


def read_file_statistics(file_name):
    stat_dict = {}
    with open(file_name, mode="r", encoding="windows-1257'") as file:
        for line in file:
            pho, freq = line.strip().split("\t")
            stat_dict[pho] = int(freq)
    return stat_dict


def normalize_feature_count(feature_count_dict, total_nr_of_feature):
    return {feature: count / total_nr_of_feature for feature, count in feature_count_dict.items()}

#def sort_dict_by_value_return_keys_as_list(feature_count_dict, reverse=True):
#    return [key for key, _ in sorted(feature_count_dict.items(), key=lambda x: x[1], reverse=reverse)]

#def sort_dict_by_value_return_keys_as_list(feature_count_dict, reverse=True):
    # Convert keys to strings to ensure proper sorting when keys are numeric
#    return [str(key) for key, _ in sorted(feature_count_dict.items(), key=lambda x: x[1], reverse=reverse)]

def sort_dict_by_value_return_keys_as_list(feature_count_dict, reverse=True):
    # Return keys in original format, sorted by values
    return [key for key, _ in sorted(feature_count_dict.items(), key=lambda x: x[1], reverse=reverse)]

def find_missing_feature(ref_feature_count_dict, focus_feature_count_dict):
    """
    Finds missing features by checking if they are absent in the focus dictionary
    or have a frequency of zero.
    """
    missing_keys = set(ref_feature_count_dict) - set(focus_feature_count_dict)
    
    # Lisa need võtmed, mis on olemas, aga sagedus on 0
    zero_freq_keys = {key for key, value in focus_feature_count_dict.items() if value == 0}

    return list(missing_keys | zero_freq_keys)


def create_table_str_for_missing_phone_s(missing_feature_List_sorted, ref_feature_count_dict, ref_feature_count_norm, label, feature_label):
    header = ["No", f"{feature_label}", "Ref corpus count ↓", "Ref corpus %"]
    title = f"\n\n- FOOKUSKORPUSES PUUDUVAD '{label}'\n\n"
    
    if not missing_feature_List_sorted:
        return title + "  ---"
    
    # Kontrolli, kas teine veerg sisaldab arve
    is_second_column_numeric = all(str(feature).replace('.', '', 1).isdigit() for feature in missing_feature_List_sorted)
    
    # Muuda joondust vastavalt sellele, kas teine veerg on arvuline või mitte
    second_column_format = ">" if is_second_column_numeric else "<"
    
    # Arvuta veergude laiused
    w1 = len(header[0])
    w2 = len(header[1])
    w3 = len(header[2])
    w4 = len(header[3])
    
    # Loo tabeli päis
    table_string = title + \
                   f"{header[0]:>{w1}} | {header[1]:{second_column_format}{w2}} | {header[2]:>{w3}} | {header[3]:>{w4}}\n" + \
                   "-" * (w1 + w2 + w3 + w4 + 9) + "\n"
    
    # Loo tabeli read
    for seq_nr, feature in enumerate(missing_feature_List_sorted, 1):
        # Kui feature puudub referents sõnastikest, lisame selle nullväärtusega
        if feature not in ref_feature_count_dict:
            ref_feature_count_dict[feature] = 0
        if feature not in ref_feature_count_norm:
            ref_feature_count_norm[feature] = 0.0
        
        ref_count = ref_feature_count_dict[feature]
        ref_percent = ref_feature_count_norm[feature] * 100
        row = f"{seq_nr:>{w1}} | {feature:{second_column_format}{w2}} | {ref_count:>{w3}} | {ref_percent:>{w4}.3f}\n"
        table_string += row.replace(",", " ")
    
    return table_string


def filter_zero_elements(ref_dict, focus_dict):
    filtered_ref_dict = {}
    filtered_focus_dict = {}
    for phoneme in ref_dict:
        if ref_dict[phoneme] > 0 or focus_dict.get(phoneme, 0) > 0:
            filtered_ref_dict[phoneme] = ref_dict[phoneme]
            filtered_focus_dict[phoneme] = focus_dict.get(phoneme, 0)
    return filtered_ref_dict, filtered_focus_dict


def find_significant_diff_feature(ref_feature_count_dict_norm, ref_total_nr_of_feature, focus_feature_count_dict_norm, focus_total_nr_of_feature):
    significant_diff_feature = {}
    for feature in set(ref_feature_count_dict_norm).intersection(focus_feature_count_dict_norm):
        ref_freq = ref_feature_count_dict_norm[feature] * ref_total_nr_of_feature
        focus_freq = focus_feature_count_dict_norm[feature] * focus_total_nr_of_feature
        
        # Kontrollime, et mõlemad sagedused on suuremad kui 0
        if ref_freq > 0 and focus_freq > 0:
            obs = np.array([
                [ref_freq, (1 - ref_feature_count_dict_norm[feature]) * ref_total_nr_of_feature],
                [focus_freq, (1 - focus_feature_count_dict_norm[feature]) * focus_total_nr_of_feature]
            ])
            chi2, p, _, _ = chi2_contingency(obs)
            if p < 0.05 and abs(focus_feature_count_dict_norm[feature]) >= 0.00001:
                significant_diff_feature[feature] = round(p, 4)
    return significant_diff_feature



def sort_features(sig_diff_feature, focus_feature_count_norm, ref_feature_count_norm):
    # Arvuta erinevused ja sorteeri need kahanevalt absoluutväärtuse järgi
    sorted_features = sorted(
        sig_diff_feature.items(),
        key=lambda x: abs(focus_feature_count_norm[x[0]] - ref_feature_count_norm[x[0]]),
        reverse=True
    )
    return sorted_features

def create_table_str_for_significantz_diff_bigams(sig_diff_feature, ref_feature_count_dict, ref_feature_count_dict_norm, focus_feature_count_dict, focus_feature_count_dict_norm, label, feature_label, threshold):
    header = ["No", f"{feature_label}", "p-value", "Ref corp count", "Ref corp %", "Focus corp count", "Focus corp %", "Focus - Ref dif % ↓"]
    
    # Kontrolli, kas teine veerg sisaldab arve
    is_second_column_numeric = all(str(feature).replace('.', '', 1).isdigit() for feature in sig_diff_feature.keys())
    
    # Muuda joondust vastavalt sellele, kas teine veerg on arvuline või mitte
    second_column_format = ">" if is_second_column_numeric else "<"
    
    # Arvuta veergude laiused
    w1 = len(header[0])
    w2 = len(header[1])
    w3 = len(header[2])
    w4 = len(header[3])
    w5 = len(header[4])
    w6 = len(header[5])
    w7 = len(header[6])
    w8 = len(header[7])
    
    # Loo tabeli päis
    table_string = f"\n\n\n- FOOKUSKORPUSE '{label}', MILLISTE NORMALISEERITUD SAGEDUSE ERINEB OLULISELT REF KORPUSE OMAST\n\n" + \
                   f"{header[0]:>{w1}} | {header[1]:{second_column_format}{w2}} | {header[2]:>{w3}} | {header[3]:>{w4}} | {header[4]:>{w5}} | {header[5]:>{w6}} | {header[6]:>{w7}} | {header[7]:>{w8}}\n" + \
                   "-" * (w1 + w2 + w3 + w4 + w5 + w6 + w7 + w8 + 21) + "\n"
    
    # Sorteeri omadused
    sorted_features = sort_features(sig_diff_feature, focus_feature_count_dict_norm, ref_feature_count_dict_norm)
    
    # Loo tabeli read
    for seq_nr, (feature, p_value) in enumerate(sorted_features, 1):
        ref_count = ref_feature_count_dict[feature]
        ref_percent = ref_feature_count_dict_norm[feature] * 100
        focus_count = focus_feature_count_dict[feature]
        focus_percent = focus_feature_count_dict_norm[feature] * 100
        dif = focus_percent - ref_percent
        print("DIF: ", dif)
        if abs(dif) >= threshold:
            row = f"{seq_nr:>{w1}} | {feature:{second_column_format}{w2}} | {p_value:>{w3}.3f} | {ref_count:>{w4},} | {ref_percent:>{w5}.3f} | {focus_count:>{w6},} | {focus_percent:>{w7}.3f} | {dif:>{w8}.3f}\n"
            table_string += row.replace(",", " ")
        else:
            break
    
    return table_string

def kolmogorov_smirnov_test(ref_feature_count_dict_norm, focus_feature_count_dict_norm, alpha):
    unique_phones = set(ref_feature_count_dict_norm).union(focus_feature_count_dict_norm)
    ref_distribution = [ref_feature_count_dict_norm.get(item, 0) for item in unique_phones]
    focus_distribution = [focus_feature_count_dict_norm.get(item, 0) for item in unique_phones]

    # Kontrollime, et jaotused ei ole tühjad
    if not any(ref_distribution) or not any(focus_distribution):
        return "\nKolmogorov-Smirnovi test jäi ära, sest vähemalt üks jaotus on tühi või sisaldab ainult nullväärtusi.\n"

    D, p_value = stats.ks_2samp(ref_distribution, focus_distribution)
    answer_string = f"\nD-statistiku väärtus: {round(D, 3)}\np-väärtus: {round(p_value, 8)}\n"
    
    if p_value < alpha:
        answer_string += f"\nKuna p-väärtus {round(p_value, 8)} < {alpha}, saame järeldada et, \nnullhüpotees tagasilükatud! St, fookuskorpuse jaotus EI ALLU referentskorpuse jaotusele!"
    else:
        answer_string += f"\nKuna p-väärtus {round(p_value, 8)} > {alpha}, saame järeldada et, \nnullhüpotees mitte-tagasilükatud! St, fookuskorpuse jaotus ALLUB referentskorpuse jaotusele!"
    
    return answer_string


def print_to_file(filename, mode, text):
    with open(filename, mode=mode, encoding="utf-8") as f_output:
        f_output.write(text)

if __name__ == "__main__":
    import traceback
    try:
        # Küsime kasutajalt fookusteksti faili ja referentsfaili asukoha:
        focus_results_file = input("Sisesta fookuskorpuse analüüsi faili täistee: ").strip()
        ref_corp_results_file = input("Sisesta referentskorpuse analüüsi faili täistee: ").strip()
        
        # Teie varasem kood:
        # -----------------------------------------
        focus_data_dict = extract_frequencies_from_file_to_dict(focus_results_file)
        ref_data_dict = extract_frequencies_from_file_to_dict(ref_corp_results_file)
        
        timestamp = datetime.now().strftime("%Y.%m.%d_%H-%M-%S")
        results_filename = f"Focus-reference_results_{timestamp}.txt"

        
        label_dict = {
            'foneemide_koguarv': 'FONEEMID', 
            'foneem_rõhulises_silbis': 'FONEEMID RÕHULISES SILBIS', 
            'foneem_rõhuta_silbis': 'FONEEMID RÕHUTA SILBIS', 
            'foneem_fraasi_sees': 'FONEEMID FRAASI SEES', 
            'foneem_fraasi_viimases_sõnas': 'FONEEMID FRAASI VIIMASES SÕNAS', 
            'diftong': 'DIFTONGID', 
            'foneeme_silbis': 'SILBISTRUKTUUR: FONEEME SILBIS',
            'silpe_sõnas': "SÕNASTRUKTUUR: SILPE SÕNAS",
            'sõnu_fraasis': 'FRAASISTTRUKTUUR: SÕNU FRAASIS',
            'sõnu_lauses': 'LAUSESTRUKTUUR: SÕNU LAUSES',
            'difoon': 'DIFOONID'
        }
        
        feat_label_dict = {
            'foneemide_koguarv': 'foneem', 
            'foneem_rõhulises_silbis': 'foneem', 
            'foneem_rõhuta_silbis': 'foneem', 
            'foneem_fraasi_sees': 'foneem', 
            'foneem_fraasi_viimases_sõnas': 'foneem', 
            'diftong': 'diftong', 
            'foneeme_silbis': 'foneeme',
            'silpe_sõnas': "silpe",
            'sõnu_fraasis': 'sõnu',
            'sõnu_lauses': 'sõnu',
            'difoon': 'difoon'
        }
        
        with open(results_filename, "w", encoding="windows-1257'") as f:
            pass
        
        for i, key in enumerate(ref_data_dict.keys()):
            label = label_dict[key]
            feature_label = feat_label_dict[key]
            
            text = (('\n' * 4) if i > 0 else '') + f"*** {label} ***\n"
            text += "=" * 117
            
            ref_feature_count_dict = ref_data_dict[key]
            focus_feature_count_dict = focus_data_dict[key]
            
            ref_total_nr_of_feature = sum(ref_feature_count_dict.values())
            focus_total_nr_of_feature = sum(focus_feature_count_dict.values())
            
            ref_feature_count_dict_norm = normalize_feature_count(ref_feature_count_dict, ref_total_nr_of_feature)
            focus_feature_count_dict_norm = normalize_feature_count(focus_feature_count_dict, focus_total_nr_of_feature)
            
            ref_feature_count_dict_norm, focus_feature_count_dict_norm = filter_zero_elements(ref_feature_count_dict_norm, focus_feature_count_dict_norm)
            
            ref_feature_sorted_list = sort_dict_by_value_return_keys_as_list(ref_feature_count_dict)
            
            missing_feature_list = find_missing_feature(ref_feature_count_dict, focus_feature_count_dict)
            missing_feature_List_sorted = sorted(missing_feature_list, key=ref_feature_sorted_list.index)
            
            missing_phone_s_table_str = create_table_str_for_missing_phone_s(missing_feature_List_sorted, ref_feature_count_dict, ref_feature_count_dict_norm, label, feature_label)
            print(missing_phone_s_table_str)
            
            threshold = 0.25
            
            significant_diff_feature = find_significant_diff_feature(ref_feature_count_dict_norm, ref_total_nr_of_feature, focus_feature_count_dict_norm, focus_total_nr_of_feature)
            
            sign_diff_phone_s_table_str = create_table_str_for_significantz_diff_bigams(significant_diff_feature, 
                                                                                        ref_feature_count_dict, 
                                                                                        ref_feature_count_dict_norm,
                                                                                        focus_feature_count_dict, 
                                                                                        focus_feature_count_dict_norm, 
                                                                                        label, feature_label, threshold)
            
            print(sign_diff_phone_s_table_str)
            
            alpha = 0.05 # otsustuspunkt p-value jaoks
            inference_str = kolmogorov_smirnov_test(ref_feature_count_dict_norm, focus_feature_count_dict_norm, alpha)
            print(inference_str)
            
            # Prindime tulemused faili
            text += missing_phone_s_table_str + sign_diff_phone_s_table_str + inference_str
            mode = "a+"
            print_to_file(results_filename, mode, text)
        
    except Exception:
        with open("error_log.txt", "w", encoding="utf-8") as f:
            f.write(traceback.format_exc())
        print("Ilmnes viga! Täpsem info failis 'error_log.txt'.")
    
    # Hoiame programmi lahti, et kasutaja jõuaks vajadusel veateadet lugeda
    input("Vajuta Enter, et sulgeda programm ...")

